import torch
from lightning import pytorch as pl
from torch import Tensor

from torchcross import models
from torchcross.cd.activations import get_prob_func
from torchcross.cd.losses import get_loss_func
from torchcross.data.task import TaskDescription
from torchcross.models.lightning import mixins


class CrossDomainLightningModule(
    models.CrossDomainModel, mixins.MeanMetricsMixin, pl.LightningModule
):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.automatic_optimization = False

    def get_metrics_and_loss(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
    ) -> tuple[tuple[Tensor, ...], Tensor]:
        (x, y), task_description = batch
        logits = self.forward(x, task_description)
        loss_func = get_loss_func(
            task_description.task_target, task_description.classes, self.device
        )
        pred_func = get_prob_func(task_description.task_target)
        loss = loss_func(logits, y)
        pred = pred_func(logits)
        return self.compute_metrics(task_description, pred, y), loss

    def compute_metrics(self, task_description, pred, y):
        raise NotImplementedError

    def training_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        self.optimizers().zero_grad()

        metric_values, loss = self.get_metrics_and_loss(batch)

        self.update_metrics("train", metric_values)

        self.manual_backward(loss)
        self.optimizers().step()

        self.log("loss/train", loss, batch_size=len(batch), prog_bar=True)
        self.log_dict(self.training_metrics)

    def validation_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        metric_values, loss = self.get_metrics_and_loss(batch)

        self.update_metrics("val", metric_values)

        self.log("loss/val", loss, batch_size=len(batch), prog_bar=True)
        self.log_dict(self.validation_metrics, prog_bar=True)

    def test_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        metric_values, loss = self.get_metrics_and_loss(batch)

        self.update_metrics("test", metric_values)

        self.log("loss/test", loss, batch_size=len(batch), prog_bar=True)
        self.log_dict(self.test_metrics)

    def configure_optimizers(self):
        return self.optimizer(self.parameters())



from functools import partial
from typing import Callable

import torch
import torchmetrics

from torchcross.data.task import TaskTarget


def get_accuracy_func(
    task_target: TaskTarget, classes: dict[int, str], device: torch.device = None
) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    """Get the accuracy function for a given task target and classes.

    Args:
        task_target: The task target.
        classes: The classes for the task.
        device: The device to create tensors needed for the accuracy
            function on.

    Returns:
        A torchmetrics functional metric or a function wrapping a
        torchmetrics functional metric for the given task target and
        classes.
    """
    match task_target:
        case TaskTarget.MULTICLASS_CLASSIFICATION:
            return partial(
                torchmetrics.functional.classification.multiclass_accuracy,
                num_classes=len(classes),
            )
        case TaskTarget.MULTILABEL_CLASSIFICATION:
            return partial(
                torchmetrics.functional.classification.multilabel_accuracy,
                num_labels=len(classes),
            )
        case TaskTarget.BINARY_CLASSIFICATION:
            return torchmetrics.functional.classification.binary_accuracy
        case TaskTarget.ORDINAL_REGRESSION:
            class_to_idx = torch.full(
                (max(classes) + 1,),
                -1,
                dtype=torch.long,
                device=device,
            )
            for i, c in enumerate(classes):
                class_to_idx[c] = i
            acc = partial(
                torchmetrics.functional.classification.multiclass_accuracy,
                num_classes=len(classes),
            )
            return lambda p, y: acc(p, class_to_idx[y])
        case target:
            raise NotImplementedError(f"Unsupported task target: {target}")


def get_auroc_func(
    task_target: TaskTarget, classes: dict[int, str], device: torch.device = None
) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    """Get the area under the ROC curve metric function for a given task
    target and classes.

    Args:
        task_target: The task target.
        classes: The classes for the task.
        device: The device to create tensors needed for the metric
            function on.

    Returns:
        A torchmetrics functional metric or a function wrapping a
        torchmetrics functional metric for the given task target and
        classes.
    """
    match task_target:
        case TaskTarget.MULTICLASS_CLASSIFICATION:
            return partial(
                torchmetrics.functional.classification.multiclass_auroc,
                num_classes=len(classes),
            )
        case TaskTarget.MULTILABEL_CLASSIFICATION:
            return partial(
                torchmetrics.functional.classification.multilabel_auroc,
                num_labels=len(classes),
            )
        case TaskTarget.BINARY_CLASSIFICATION:
            return torchmetrics.functional.classification.binary_auroc
        case TaskTarget.ORDINAL_REGRESSION:
            class_to_idx = torch.full(
                (max(classes) + 1,),
                -1,
                dtype=torch.long,
                device=device,
            )
            for i, c in enumerate(classes):
                class_to_idx[c] = i
            auroc = partial(
                torchmetrics.functional.classification.multiclass_auroc,
                num_classes=len(classes),
            )
            return lambda p, y: auroc(p, class_to_idx[y])
        case target:
            raise NotImplementedError(f"Unsupported task target: {target}")



import logging
from typing import Callable, Any

import torch
from torch import nn
from torch.optim import Optimizer

from torchcross import models
from torchcross.cd.metrics import get_accuracy_func, get_auroc_func
from torchcross.data.task import TaskDescription
from torchcross.models.lightning.cross_domain import CrossDomainLightningModule


class SimpleCrossDomainClassifier(
    models.SimpleCrossDomainClassifier, CrossDomainLightningModule
):
    def __init__(
        self,
        backbone: tuple[torch.nn.Module, int],
        optimizer: type[Optimizer],
        lr_scheduler: Callable[[Optimizer], Any] | None = None,
        task_descriptions: list[TaskDescription] = None,
        add_heads_during_training: bool = True,
    ):
        super().__init__(
            *backbone,
            task_descriptions,
            add_heads_during_training,
        )

        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

        metric_keys = ["accuracy/{}", "AUROC/{}"]
        self.configure_metrics(metric_keys)

    def add_head(self, task_description: TaskDescription) -> nn.Module:
        new_head = super().add_head(task_description)
        self.optimizers().add_param_group({"params": new_head.parameters()})
        return new_head

    def compute_metrics(self, task_description, pred, y):
        accuracy_func = get_accuracy_func(
            task_description.task_target, task_description.classes, self.device
        )
        auroc_func = get_auroc_func(
            task_description.task_target, task_description.classes, self.device
        )
        accuracy = accuracy_func(pred, y)
        auroc = auroc_func(pred, y)
        return accuracy, auroc

    def validation_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        try:
            super().validation_step(batch, batch_idx)
        except ValueError as e:
            if "Head for task" in str(e):
                logging.warning(f"{e} Skipping validation step.")
            else:
                raise e



import bisect
import random
from abc import ABC
from collections.abc import Callable, Sequence

import numpy as np
import torch
from torch.utils.data import (
    Dataset,
    ConcatDataset,
    BatchSampler,
    RandomSampler,
    SequentialSampler,
    IterableDataset,
)

from torchcross.utils.collate_fn import identity
from .task import TaskTarget, TaskDescription

__all__ = [
    "TaskSource",
    "WrapTaskSource",
    "BatchedTaskSource",
    "RandomChainTaskSource",
    "ConcatTaskSource",
]

from ..utils import to_numpy


class TaskSource(Dataset, ABC):
    """A dataset that can be used as a source to generate tasks."""

    task_target: TaskTarget
    labels: np.ndarray | torch.Tensor | Sequence
    classes: dict[int, str]
    task_identifier: str = ""


class WrapTaskSource(TaskSource):
    """Wraps a dataset as a task source."""

    def __init__(
        self,
        dataset: Dataset,
        task_target: TaskTarget,
        classes: dict[int, str],
        label_fn: Callable[[Dataset], np.ndarray | torch.Tensor | Sequence] = None,
        task_identifier: str = "",
    ):
        self.dataset = dataset
        self.task_target = task_target
        self.classes = classes
        self.task_identifier = task_identifier

        def default_label_fn(ds: Dataset):
            return np.stack([to_numpy(item[1]) for item in ds])

        label_fn = label_fn if label_fn else default_label_fn
        self.labels = label_fn(self.dataset)

    def __getitem__(self, index):
        return self.dataset[index]

    def __len__(self):
        return len(self.dataset)


class BatchedTaskSource(TaskSource, IterableDataset):
    def __init__(
        self,
        task_source: TaskSource,
        batch_size: int,
        shuffle: bool = True,
        drop_last: bool = False,
        collate_fn=None,
        with_task_description: bool = False,
    ):
        self.task_source = task_source
        self.task_target = task_source.task_target
        self.classes = task_source.classes
        self.task_identifier = task_source.task_identifier
        self.with_task_description = with_task_description
        if self.with_task_description:
            self.task_description = TaskDescription(
                self.task_target, self.classes, self.task_identifier
            )

        sampler = (
            RandomSampler(self.task_source)
            if shuffle
            else SequentialSampler(self.task_source)
        )
        self.batch_sampler = BatchSampler(sampler, batch_size, drop_last)
        self.collate_fn = collate_fn if collate_fn else identity

    def __iter__(self):
        for batch in self.batch_sampler:
            if self.with_task_description:
                yield (
                    self.collate_fn([self.task_source[i] for i in batch]),
                    self.task_description,
                )
            else:
                yield self.collate_fn([self.task_source[i] for i in batch])

    def __len__(self):
        return len(self.batch_sampler)


class RandomChainTaskSource(TaskSource, IterableDataset):
    def __init__(self, task_sources: list[BatchedTaskSource]):
        self.task_sources = task_sources

    def __iter__(self):
        iterators = [iter(ts) for ts in self.task_sources]
        # Calculate the weights of each iterator based on the batched task source length
        weights = [len(ts) for ts in self.task_sources]
        task_descriptions = [
            TaskDescription(ts.task_target, ts.classes, ts.task_identifier)
            for ts in self.task_sources
        ]
        while iterators:
            # Randomly select an iterator with a probability proportional to its weight
            i = random.choices(range(len(self.task_sources)), weights=weights)[0]
            selected_iterator = iterators[i]
            try:
                # Return one element from the selected iterator
                yield next(selected_iterator), task_descriptions[i]
            except StopIteration:
                # The iterator is exhausted, so remove it from the list of iterators
                iterators.remove(selected_iterator)

    def __len__(self):
        return sum(len(ts) for ts in self.task_sources)


class ConcatTaskSource(TaskSource, ConcatDataset):
    datasets: list[TaskSource]

    def __getitem__(self, idx):
        if idx < 0:
            if -idx > len(self):
                raise ValueError(
                    "absolute value of index should not exceed dataset length"
                )
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return (
            self.datasets[dataset_idx][sample_idx],
            TaskDescription(
                self.datasets[dataset_idx].task_target,
                self.datasets[dataset_idx].classes,
                self.datasets[dataset_idx].task_identifier,
            ),
        )

    def __len__(self):
        return sum(len(d) for d in self.datasets)


class SubsetTaskSource(TaskSource):
    def __init__(self, task_source: TaskSource, indices: Sequence[int]):
        self.task_source = task_source
        self.indices = indices
        self.task_target = task_source.task_target
        self.classes = task_source.classes
        self.task_identifier = task_source.task_identifier
        self.labels = task_source.labels[indices]

    def __getitem__(self, idx):
        return self.task_source[self.indices[idx]]

    def __len__(self):
        return len(self.indices)



from collections.abc import Collection
from dataclasses import dataclass
from enum import Enum, auto

from torch.utils.data import Dataset


class TaskTarget(Enum):
    MULTICLASS_CLASSIFICATION = auto()
    BINARY_CLASSIFICATION = auto()
    MULTILABEL_CLASSIFICATION = auto()
    ORDINAL_REGRESSION = auto()
    REGRESSION = auto()
    BINARY_SEGMENTATION = auto()
    MULTILABEL_SEGMENTATION = auto()


@dataclass
class Task:
    """A task with support and query datasets as well as a task target
    and classes.

    Args:
        support: The support dataset or collection of inputs and targets.
        query: The query dataset or collection of inputs and targets.
        task_target: The task target.
        classes: The classes for the task.
    """

    support: Dataset | Collection
    query: Dataset | Collection
    task_target: TaskTarget
    classes: dict[int, str] = None


@dataclass
class TaskDescription:
    """A task description with a task target and classes as well as an
    optional string identifier for the task.

    Args:
        task_target: The task target.
        classes: The classes for the task.
        task_identifier: The task identifier.
    """

    task_target: TaskTarget
    classes: dict[int, str] = None
    task_identifier: str = ""



import math
from collections import OrderedDict
from collections.abc import Sequence, Callable
from typing import Optional, Literal, Tuple, List

import torch
import torchopt
from torch import nn, Tensor

from torchcross.cd.heads import new_classification_head
from torchcross.cd.losses import get_loss_func
from torchcross.data.metadataset.few_shot import get_indices
from torchcross.data.task import TaskDescription, Task, TaskTarget
from torchcross.models.episodic import EpisodicModel


class ProtoNet(EpisodicModel):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        distance: Literal["euclidean", "cosine", "squared_euclidean"] = "euclidean",
    ) -> None:
        super(ProtoNet, self).__init__()

        self.backbone = backbone
        self.num_backbone_features = num_backbone_features

        self.distance = distance

    def episode(self, task: Task) -> torch.Tensor:
        prototypes, neg_prototypes = self.get_prototypes(task)
        query_x = task.query[0]
        query_features = self.backbone(query_x)
        query_logits = torch.cdist(query_features, prototypes)

        return query_logits

    def get_prototypes(self, task: Task) -> tuple[Tensor, Tensor]:
        support_x, support_y = task.support
        support_features = self.backbone(support_x)
        task_target = task.task_target
        classes = task.classes

        if task_target is TaskTarget.MULTICLASS_CLASSIFICATION:
            prototypes = torch.stack(
                [support_features[support_y == c].mean(dim=0) for c in classes]
            )
            neg_prototypes = None
        elif task_target is TaskTarget.MULTILABEL_CLASSIFICATION:
            prototypes = torch.stack(
                [support_features[support_y[:, c] == 1].mean(dim=0) for c in classes]
            )
            neg_prototypes = torch.stack(
                [support_features[support_y[:, c] == 0].mean(dim=0) for c in classes]
            )
        elif task_target is TaskTarget.BINARY_CLASSIFICATION:
            prototypes = torch.stack([support_features[support_y == 1].mean(dim=0)])
            neg_prototypes = torch.stack([support_features[support_y == 0].mean(dim=0)])
        elif task_target is TaskTarget.ORDINAL_REGRESSION:
            raise NotImplementedError("Ordinal regression not yet implemented")
        else:
            raise NotImplementedError(f"Task target {task_target} not yet implemented")
        return prototypes, neg_prototypes
